// -*- C++ -*-

#pragma once

#include "lib/range.h"
#include "proto.h"

namespace goopax {
namespace impl {

/**
     \ingroup types
     \brief type info and type conversion mechanism
       With the gettype struct, CPU and GPU types can be converted to each other, and type information obtained
       typename gettype<T>::cpu converts T to the corresponding CPU type. T can be CPU type, GPU type, or CPU debug type, or a goopax struct type.
       typename gettype<T>::gpu converts T to the corresponding GPU type. T can be CPU type, GPU type, or CPU debug type, or a goopax struct type.
       gettype<T>::size provides the size of type T in bytes. T can be CPU type, GPU type, or CPU debug type, or a goopax struct type.
       typename gettype<T>::template change<U>::type changes the type of T to type U while preserving CPU/GPU/debug mode. If T is CPU type, the resulting type is also a CPU type. If T is GPU type, the resulting type is a GPU type. If T is debug data type, the resulting type is also a debug data type.
     */
template<typename T, typename EIF = void>
struct gettype;


template<typename T>
struct gettype_base
{
    static_assert(is_arithmetic<T>::value || is_pointer<T>::value, "");
    using cpu                    = goopax::impl::checknantype<T>;
    using gpu                    = gpu_type<T>;
    static constexpr size_t size = sizeof(T);
};
#if GOOPAX_CONSTEXPR_INSTANTIATION
template<typename T>
constexpr size_t gettype_base<T>::size;
#endif

template<typename T>
struct gettype_base<T*>
{
    using cpu                    = T*;
    using gpu                    = gpu_type<T*>;
    static constexpr size_t size = sizeof(void*);
};
#if GOOPAX_CONSTEXPR_INSTANTIATION
template<typename T>
constexpr size_t gettype_base<T*>::size;
#endif

template<typename T>
struct gettype_base<const T*>
{
    using cpu                    = const T*;
    using gpu                    = gpu_type<const T*>;
    static constexpr size_t size = sizeof(void*);
};
#if GOOPAX_CONSTEXPR_INSTANTIATION
template<typename T>
constexpr size_t gettype_base<const T*>::size;
#endif


template<typename T>
struct gettype<gpu_type<T*>> : public gettype_base<T*>
{
    using gpu = gpu_type<T*>;

    template<typename U>
    struct change
    {
        using type = gpu_type<U>;
    };
};

template<typename T>
struct gettype<gpu_type<T>> : public gettype_base<T>
{
    template<typename U>
    struct change
    {
        using type = gpu_type<typename unrangetype_core<U>::type>;
    };
};

template<typename T, typename TA>
struct gettype<checknan<T, TA>> : public gettype_base<T>
{
    template<typename U>
    struct change
    {
        using type = checknan<typename unrangetype_core<U>::type, TA>;
    };
};

template<typename TA>
struct gettype<checknan<xtype_int, TA>>
{
    template<typename U>
    struct change
    {
        using type = checknan<typename unrangetype_core<U>::type, TA>;
    };
};
template<>
struct gettype<xtype_int>
{
    template<typename U>
    struct change
    {
        using type = typename unrangetype_core<U>::type;
    };
};


template<typename T, bool ARITH>
struct gettype_split;

template<typename STR>
struct gettype_split<STR, false>
{
    // defaults to goopax struct type
    using cpu                    = typename goopax_struct_changetype<typename remove_cv<STR>::type, goopax::impl::Tulong>::type;
    using gpu                    = typename goopax_struct_changetype<typename remove_cv<STR>::type, gpu_type<unsigned long>>::type;
    static constexpr size_t size = sizeof(typename goopax_struct_changetype<typename remove_cv<STR>::type, unsigned long>::type);
    template<typename U>
    struct change
    {
        using type = typename gettype<typename get_X_type<STR>::type>::template change<U>::type;
    };
};
#if GOOPAX_CONSTEXPR_INSTANTIATION
template<typename STR>
constexpr size_t gettype_split<STR, false>::size;
#endif

template<typename T>
struct gettype_split<T, true> : public gettype_base<T>
{
    template<typename U>
    struct change
    {
        using type = typename unrangetype<U>::type;
    };
};


template<typename T>
struct gettype<T, typename enable_if<(is_arithmetic<T>::value || is_pointer<T>::value || is_gpu_struct<T>::value) && !is_checknan<T>::value>::type> : public gettype_split<T, is_arithmetic<T>::value || is_pointer<T>::value>
{};


template<typename T, size_t N>
struct gettype<gpu_struct::struct_mod_test<T, N>>
{
    template<typename U>
    struct change
    {
        using type = gpu_struct::struct_mod_test<U, N>;
    };
    using cpu = gpu_struct::struct_mod_test<typename gettype<T>::cpu, N>;
    using gpu = gpu_struct::struct_mod_test<typename gettype<T>::gpu, N>;
};

template<typename T>
struct gettype<gpu_struct::struct_item_map<T>>
{
    template<typename U>
    struct change
    {
        using type = gpu_struct::struct_item_map<typename gettype<T>::template change<U>::type>;
    };
    using cpu = gpu_struct::struct_item_map<typename gettype<T>::cpu>;
    using gpu = gpu_struct::struct_item_map<typename gettype<T>::gpu>;
};


}
}
