// -*- C++ -*-

#pragma once

#include "memory.h"

/**
\addtogroup loops
Loops in gpu kernels can be achieved by the use of gpu_for or gpu_while statements.
Alternatively, small loops that don't depend on gpu variables may also be implemented as standard C++ for/while loops. This will lead to explicit unrolling of the loop.

The usage of the gpu_for loop is different from C++ for loops and looks like this:
\code{.cpp}
gpu_for(0, 10, [&](gpu_uint k)
{
  <do something> // k runs from 0 to 9
});
\endcode

Step sizes can be provided as a third argumend:
 \code{.cpp}
 gpu_for(0, 10, 2, [&](gpu_int k)
 {
    <do something> // only runs over even numbers form 0 to 8
 });
 \endcode
 
 The comparison operator can be specified as a template argument (defaults to std::less<>):
 \code{.cpp}
 gpu_for<std::less_equal<>>(0, 10, 2, [&](gpu_int k)
 {
    <do something> // only runs over even numbers form 0 to 10
 });
 \endcode

\{
*/


#if defined(__ENVIRONMENT_IPHONE_OS_VERSION_MIN_REQUIRED__) && (__ENVIRONMENT_IPHONE_OS_VERSION_MIN_REQUIRED__ < 90000)
#define GOOPAX_HAVE_CPU_IF 0
#else
#define GOOPAX_HAVE_CPU_IF 1
#endif


namespace goopax {
namespace impl {
namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) gpu_if_impl* goopax_4_if_create(const char*, unsigned int, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) gpu_if_impl* goopax_4_elseif_create(const char*, unsigned int, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_infloop_set_next(loop_impl*, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) bool goopax_4_infloop_do_continue(loop_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) loop_impl* goopax_4_infloop_create(const char*, unsigned int, gpu_writer_base_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void* goopax_4_loop_access() noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_loop_make(gpu_writer_base_impl*, const gpu_writer_base_impl*, const gpu_writer_base_impl*, int) noexcept;
}
}

template<typename P, memory::scope scope>
class gpu_ptrbase;

struct elsedata
{
#ifdef NDEBUG
    using Tbool = bool;
#else
    using Tbool = goopax::debug::types::Tbool;
#endif
#if GOOPAX_HAVE_CPU_IF
    Tbool know;
    Tbool value;
#endif
};

#if GOOPAX_HAVE_CPU_IF
static thread_local elsedata Elsedata;
#endif

struct gpu_if_clause
{
    using Tbool = elsedata::Tbool;

    private:
#if GOOPAX_HAVE_CPU_IF
    elsedata&   ed;
    const Tbool know;
    const Tbool prev_else_value;
    const Tbool value;
#endif

    public:
    int loopvar = 0;

    unique_ptr<gpu_if_impl, interface::goopax_deleter> impl;

#if GOOPAX_HAVE_CPU_IF
    inline gpu_if_clause(elsedata& ed0, const char*, unsigned int, false_type, bool cnd)
        : ed(ed0)
        , know(true)
        , prev_else_value(true)
        , value(cnd)
        , loopvar(cnd ? 0 : 1)
    {
        ed = elsedata();
    }
    inline gpu_if_clause(elsedata& ed0, const char* FILE, unsigned int LINE, true_type, bool cnd)
        : ed(ed0)
        , know(ed.know)
        , prev_else_value(ed.value)
        , value(know ? Tbool(cnd && prev_else_value) : Tbool())
        , loopvar(!know || value ? 0 : 1)
    {
        if (!know)
        {
            gpu_bool cnd2 = cnd;
            impl.reset(interface::goopax_4_elseif_create(FILE, LINE, cnd2.impl.get()));
        }
        ed = elsedata();
    }
#else
    template<typename BAD = void>
    inline gpu_if_clause(const char*, unsigned int, false_type, bool)
    {
        static_assert(!is_same<BAD, void>::value, "CPU version of gpu_if requires IOS >= 9");
    }
#endif

    inline gpu_if_clause(
#if GOOPAX_HAVE_CPU_IF
        elsedata& ed0,
#endif
        const char* FILE, unsigned int LINE, false_type, const gpu_bool& cnd, ...)
        :
#if GOOPAX_HAVE_CPU_IF
        ed(ed0)
        , know(false)
        , prev_else_value(Tbool())
        , value(Tbool())
        ,
#endif
        impl(interface::goopax_4_if_create(FILE, LINE, cnd.impl.get()))
    {
#if GOOPAX_HAVE_CPU_IF
        ed = elsedata();
#endif
    }
    inline gpu_if_clause(
#if GOOPAX_HAVE_CPU_IF
        elsedata& ed0,
#endif
        const char* FILE, unsigned int LINE, true_type, const gpu_bool& cnd, ...)
#if GOOPAX_HAVE_CPU_IF
        : ed(ed0)
        , know(false)
        , prev_else_value(Tbool())
        , value(Tbool())
#endif
    {
#if GOOPAX_HAVE_CPU_IF
        if (ed.know)
        {
            gpu_if_clause oldelse(ed, FILE, LINE, false_type(), !ed.value);
            (void)oldelse;
        }
#endif
        impl.reset(interface::goopax_4_elseif_create(FILE, LINE, cnd.impl.get()));
#if GOOPAX_HAVE_CPU_IF
        ed = elsedata();
#endif
    }
    inline ~gpu_if_clause()
    {
#if GOOPAX_HAVE_CPU_IF
        ed.know = know;
        if (know)
        {
            ed.value = prev_else_value && !value;
        }
        else
        {
            ed.value = Tbool();
        }
#endif
    }
    gpu_if_clause(const gpu_if_clause&) = delete;
    gpu_if_clause& operator=(const gpu_if_clause&) = delete;
};


inline void gpu_break()
{
    check_errno(interface::goopax_4E_loop_break(nullptr));
}
inline void gpu_continue()
{
    check_errno(interface::goopax_4E_loop_continue(nullptr));
}


struct infinite_loop
{
    protected:
    unique_ptr<loop_impl, interface::goopax_deleter> impl;

    public:
    inline void loop_set_next(const gpu_bool& cont)
    {
        interface::goopax_4_infloop_set_next(impl.get(), cont.impl.get());
    }
    inline goopax::impl::Tbool loop_do_continue() const
    {
        return interface::goopax_4_infloop_do_continue(impl.get());
    }

    inline infinite_loop(const char* FILE, unsigned int LINE, const gpu_bool& initial)
        : impl(interface::goopax_4_infloop_create(FILE, LINE, initial.impl.get()))
    {}

    inline ~infinite_loop() = default;
};


#if GOOPAX_HAVE_CPU_IF
/// \ingroup loops
/// gpu_if, to be used in kernels. Takes gpu_bool as input. The block following the gpu_if() statement is executed by all threads for which the argument is true.
#define gpu_if(...) for (::goopax::impl::gpu_if_clause __goopax_if_obj(::goopax::impl::Elsedata, __FILE__, __LINE__, false_type(), __VA_ARGS__); __goopax_if_obj.loopvar < 1; ++__goopax_if_obj.loopvar)
/// \ingroup loops
/// gpu_elseif, must follow a gpu_if or gpu_elseif block. The block following the gpu_elseif() statement is executed by all threads for which the argument is true.
#define gpu_elseif(...) for (::goopax::impl::gpu_if_clause __goopax_if_obj(::goopax::impl::Elsedata, __FILE__, __LINE__, true_type(), __VA_ARGS__); __goopax_if_obj.loopvar < 1; ++__goopax_if_obj.loopvar)
#else
#define gpu_if(...) for (::goopax::impl::gpu_if_clause __goopax_if_obj(__FILE__, __LINE__, false_type(), __VA_ARGS__); __goopax_if_obj.loopvar < 1; ++__goopax_if_obj.loopvar)
#define gpu_elseif(...) for (::goopax::impl::gpu_if_clause __goopax_if_obj(__FILE__, __LINE__, true_type(), __VA_ARGS__); __goopax_if_obj.loopvar < 1; ++__goopax_if_obj.loopvar)
#endif

/// \ingroup loops
/// gpu_elseif must follow a gpu_if or gpu_elseif block. The block following the gpu_elseif statement is executed by all threads that did not execute the previous if blocks.
#define gpu_else gpu_elseif(true)

/// \ingroup loops
/// while loop.
#define gpu_while(...) for (::goopax::impl::infinite_loop LOOP(__FILE__, __LINE__, __VA_ARGS__); LOOP.loop_do_continue(); LOOP.loop_set_next(__VA_ARGS__))


template<typename T>
gpu_type<T> get_loop_type(gpu_type<T>, ...);


struct loop_ptr_converter
{
    void* ptr;
    template<typename T>
    operator T*()
    {
        return static_cast<T*>(ptr);
    }
};


template<typename T>
static constexpr int comp_enum = -1;
template<typename T>
static constexpr int comp_enum<std::less<T>> = 0;
template<typename T>
static constexpr int comp_enum<std::less_equal<T>> = 1;
template<typename T>
static constexpr int comp_enum<std::greater<T>> = 2;
template<typename T>
static constexpr int comp_enum<std::greater_equal<T>> = 3;
template<typename T>
static constexpr int comp_enum<std::equal_to<T>> = 4;
template<typename T>
static constexpr int comp_enum<std::not_equal_to<T>> = 5;


template<typename F, typename Ret, typename A, typename... Rest>
A get_type_func(Ret (F::*)(A, Rest...));

template<typename F, typename Ret, typename A, typename... Rest>
A get_type_func(Ret (F::*)(A, Rest...) const);

template<typename F>
struct first_argument
{
    using type = decltype(get_type_func(&F::operator()));
};
template<typename F>
struct step_type
{
    using T    = typename first_argument<F>::type;
    using type = decltype(declval<T>() - declval<T>());
};

template<typename T>
struct get_pointer_step;

template<typename T, memory::scope scope>
struct get_pointer_step<gpu_type<T, scope>>
{
    static_assert(!is_pointer<T>::value);
    static constexpr unsigned int value = 1;
};

template<typename T, memory::scope scope>
struct get_pointer_step<gpu_type<T*, scope>>
{
    static constexpr unsigned int value = gpu_ptrbase<T*, scope>::base_step;
};

template<typename T, memory::scope scope>
struct get_pointer_step<gpu_type<const T*, scope>> : public get_pointer_step<gpu_type<T*, scope>>
{};


template<typename comp_t, typename T, typename S, typename func_t>
inline void gpu_for_split(T begin, T end, S step, func_t func, true_type)
{
    static_assert(comp_enum<comp_t> != -1);
    typename first_argument<func_t>::type a     = begin;
    const S                               step2 = step * get_pointer_step<T>::value;
    check_errno(interface::goopax_4E_loop_make(a.impl.get(), end.impl.get(), step2.impl.get(), comp_enum<comp_t>));
    func(a);
}

template<typename comp_t, typename T, typename S, typename func_t>
inline void gpu_for_split(T begin, T end, S step, func_t func, false_type)
{
    for (typename first_argument<func_t>::type a = begin; comp_t()(a, end); a += step)
    {
        func(a);
    }
}


/// \brief gpu for loop.
/// \ingroup loops
template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for(typename first_argument<func_t>::type begin,
                    typename first_argument<func_t>::type end,
                    typename step_type<func_t>::type      step,
                    func_t                                func)
{
    constexpr bool is_gpu = is_gpu_type<typename first_argument<func_t>::type>::value;
    gpu_for_split<comp_t>(begin, end, step, func, integral_constant<bool, is_gpu>());
}

template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for(typename first_argument<func_t>::type begin,
                    typename first_argument<func_t>::type end,
                    func_t                                func)
{
    gpu_for<comp_t>(begin, end, 1, func);
}


/// \ingroup loops
/// gpu loop, parallelized over threads in workgroup.
template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for_local(typename first_argument<func_t>::type begin,
                          typename first_argument<func_t>::type end,
                          typename step_type<func_t>::type      step,
                          func_t                                func)
{
    constexpr bool is_gpu = is_gpu_type<typename first_argument<func_t>::type>::value;
    gpu_for<comp_t>(begin + (is_gpu ? local_id() * step : 0),
                    end,
                    step * (is_gpu ? local_size() : 1u),
                    func);
}

template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for_local(typename first_argument<func_t>::type begin,
                          typename first_argument<func_t>::type end,
                          func_t                                func)
{
    gpu_for_local<comp_t>(begin, end, 1, func);
}


/// \ingroup loops
/// gpu loop, parallelized over all threads
template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for_global(typename first_argument<func_t>::type begin,
                           typename first_argument<func_t>::type end,
                           typename step_type<func_t>::type      step,
                           func_t                                func)
{
    constexpr bool is_gpu = is_gpu_type<typename first_argument<func_t>::type>::value;
    gpu_for<comp_t>(begin + (is_gpu ? global_id() * step : 0),
                    end,
                    step * (is_gpu ? global_size() : 1u),
                    func);
}

template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for_global(typename first_argument<func_t>::type begin,
                           typename first_argument<func_t>::type end,
                           func_t                                func)
{
    gpu_for_global<comp_t>(begin, end, 1, func);
}


/// \ingroup loops
/// gpu loop, parallelized over workgroups
template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for_group(typename first_argument<func_t>::type begin,
                          typename first_argument<func_t>::type end,
                          typename step_type<func_t>::type      step,
                          func_t                                func)
{
    constexpr bool is_gpu = is_gpu_type<typename first_argument<func_t>::type>::value;
    gpu_for<comp_t>(begin + (is_gpu ? group_id() * step : 0),
                    end,
                    step * (is_gpu ? num_groups() : 1u),
                    func);
}

template<typename comp_t = std::less<void>, typename func_t>
inline void gpu_for_group(typename first_argument<func_t>::type begin,
                          typename first_argument<func_t>::type end,
                          func_t                                func)
{
    gpu_for_group<comp_t>(begin, end, 1, func);
}


template<typename comp_t = std::less<void>, typename T, memory::scope MEMSCOPE, typename FUNC>
void for_each(gpu_type<T*, MEMSCOPE> begin, gpu_type<T*, MEMSCOPE> end, FUNC func)
{
    gpu_for<comp_t>(begin, end, [&func](gpu_type<T*, MEMSCOPE> p) {
        func(*p);
    });
}

template<typename comp_t = std::less<void>, typename T, memory::scope MEMSCOPE, typename FUNC>
void for_each_local(gpu_type<T*, MEMSCOPE> begin, gpu_type<T*, MEMSCOPE> end, FUNC func)
{
    gpu_for_local<comp_t>(begin, end, [&func](gpu_type<T*, MEMSCOPE> p) {
        func(*p);
    });
}

template<typename comp_t = std::less<void>, typename T, memory::scope MEMSCOPE, typename FUNC>
void for_each_global(gpu_type<T*, MEMSCOPE> begin, gpu_type<T*, MEMSCOPE> end, FUNC func)
{
    gpu_for_global<comp_t>(begin, end, [&func](gpu_type<T*, MEMSCOPE> p) {
        func(*p);
    });
}

template<typename comp_t = std::less<void>, typename T, memory::scope MEMSCOPE, typename FUNC>
void for_each_group(gpu_type<T*, MEMSCOPE> begin, gpu_type<T*, MEMSCOPE> end, FUNC func)
{
    gpu_for_group<comp_t>(begin, end, [&func](gpu_type<T*, MEMSCOPE> p) {
        func(*p);
    });
}

template<typename... ARGS>
inline void for_each_local(ARGS&&... args)
{
    std::for_each(args...);
}
template<typename... ARGS>
inline void for_each_global(ARGS&&... args)
{
    std::for_each(args...);
}
template<typename... ARGS>
inline void for_each_group(ARGS&&... args)
{
    std::for_each(args...);
}

}

using impl::for_each;
using impl::for_each_global;
using impl::for_each_group;
using impl::for_each_local;
using impl::gpu_break;
using impl::gpu_continue;
using impl::gpu_for;
using impl::gpu_for_global;
using impl::gpu_for_group;
using impl::gpu_for_local;

}

//namespace std {
//using goopax::for_each;
//}


/// \}
