// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "struct.h"

namespace goopax {
namespace impl {

template<typename P, memory::scope scope>
class gpu_ptrbase : public wrapper<gpu_type<P, scope>, gpu_writer_base_impl>
{
    using TT = typename remove_cv<typename remove_pointer<P>::type>::type;

    public:
    using size_type      = typename conditional<scope == memory::device, uint64_t, uint32_t>::type;
    using ssize_type     = typename make_signed<size_type>::type;
    using gpu_size_type  = typename gettype<size_type>::gpu;
    using gpu_ssize_type = typename gettype<ssize_type>::gpu;
    using value_type     = typename goopax_struct_changetype<TT, xtype_gpu>::type;
    using basetype       = typename buffer<TT>::base_value_type;

    protected:
    static constexpr unsigned int base_step = gettype<TT>::size / gettype<basetype>::size;
    template<typename T>
    friend struct get_pointer_step;


    using gpu_P           = gpu_type<P, scope>;
    using reference       = typename conditional<std::is_const<typename remove_pointer<P>::type>::value, const value_type, value_type>::type&;
    using const_reference = const value_type&;
    using maptype         = mapframe<typename goopax_struct_changetype<TT, xtype_int>::type>;
    using gpu_maptype     = mapframe<typename goopax_struct_changetype<TT, xtype_gpu>::type>;

    public:
    gpu_P& operator+=(const gpu_ssize_type& b)
    {
        gpu_ssize_type b2 = b * base_step;
        interface::goopax_4_gpu_type_op2(6, this->impl.get(), this->impl.get(), b2.impl.get());
        return static_cast<gpu_P&>(*this);
    }
    gpu_P& operator-=(const gpu_ssize_type& b)
    {
        return static_cast<gpu_P&>(*this) += -b;
    }
    gpu_P operator+(gpu_ssize_type b) const
    {
        gpu_P ret = static_cast<const gpu_P&>(*this);
        ret += b;
        return ret;
    }
    friend gpu_P operator+(gpu_ssize_type a, gpu_P b)
    {
        return b + a;
    }

    gpu_ssize_type operator-(gpu_P b) const
    {
        gpu_ssize_type ret;
        interface::goopax_4_gpu_type_op2(18, ret.impl.get(), this->impl.get(), b.impl.get());
        return ret / base_step;
    }
    gpu_P operator-(gpu_ssize_type b) const
    {
        return (*this) + (-b);
    }

    inline gpu_P& operator++()
    {
        /// increment
        return *this += 1;
    }
    inline gpu_P& operator--()
    {
        /// decrement
        return *this -= 1;
    }
    inline gpu_P operator++(int)
    {
        /// increment
        gpu_P ret = static_cast<const gpu_P&>(*this);
        ++*this;
        return ret;
    }
    inline gpu_P operator--(int)
    {
        /// decrement
        gpu_P ret = static_cast<const gpu_P&>(*this);
        --*this;
        return ret;
    }

    gpu_bool operator<(gpu_P b) const
    {
        gpu_bool ret;
        interface::goopax_4_gpu_type_op2(2, ret.impl.get(), this->impl.get(), b.impl.get());
        return ret;
    }

    gpu_bool operator>=(gpu_P b) const
    {
        gpu_bool ret;
        interface::goopax_4_gpu_type_op2(3, ret.impl.get(), this->impl.get(), b.impl.get());
        return ret;
    }

    gpu_bool operator<=(gpu_P b) const
    {
        return b >= static_cast<const gpu_P&>(*this);
    }
    gpu_bool operator>(gpu_P b) const
    {
        return b < static_cast<const gpu_P&>(*this);
    }

    reference operator*() const
    {
        value_type* ret = new value_type;
        interface::goopax_4_register_garbage(ret, garbage_deleter<value_type>);
        interface::goopax_4_sitem_set_memref(maptype::get_static_mf(), gpu_maptype::get_static_mf(), *static_cast<const gpu_P&>(*this).impl, nullptr, gettype<basetype>::size, ret);
        return *ret;
    }

    reference operator[](gpu_ssize_type k) const
    {
        return *(static_cast<const gpu_P&>(*this) + k);
    }


    typename remove_reference<reference>::type* operator->() const
    {
        return &**this;
    }


    gpu_ptrbase()
    {
        this->impl.reset(interface::goopax_4_gpu_type_uninit(type_enum<basetype*>::value, scope));
    }
};


template<typename P, memory::scope scope>
gpu_bool operator==(const gpu_ptrbase<P, scope>& a, const gpu_ptrbase<P, scope>& b)
{
    gpu_bool ret;
    interface::goopax_4_gpu_type_op2(0, ret.impl.get(), a.impl.get(), b.impl.get());
    return ret;
}
template<typename P, memory::scope scope>
gpu_bool operator!=(const gpu_ptrbase<P, scope>& a, const gpu_ptrbase<P, scope>& b)
{
    gpu_bool ret;
    interface::goopax_4_gpu_type_op2(1, ret.impl.get(), a.impl.get(), b.impl.get());
    return ret;
}


template<typename T, memory::scope scope>
class gpu_type<T*, scope> : public gpu_ptrbase<T*, scope>
{
    using base_t = gpu_ptrbase<T*, scope>;
    using typename base_t::basetype;
    using typename base_t::gpu_maptype;
    using typename base_t::maptype;
    using typename base_t::reference;

    public:
    using typename base_t::value_type;

    template<typename EA = void, typename = typename enable_if<scope == memory::device, EA>::type>
    gpu_type(T* ptr)
    {
        this->impl.reset(interface::goopax_4_gpu_type_assign(&ptr, type_enum<basetype*>::value));
        /// conversion from CPU type
    }
    gpu_type(const gpu_type& b)
    {
        /// copy constructor
        *this = b;
    }
    gpu_type()
    {}
    gpu_type& operator=(const gpu_type& b)
    {
        /// assignment operator
        interface::goopax_4_gpu_type_copyassign(this->impl.get(), b.impl.get());
        return *this;
    }

    template<typename EA = void, typename = typename enable_if<scope == memory::device, EA>::type>
    gpu_type& operator=(T* ptr)
    {
        return *this = gpu_type(ptr);
    }
};


template<typename T, memory::scope scope>
class gpu_type<const T*, scope> : public gpu_ptrbase<const T*, scope>
{
    using base_t = gpu_ptrbase<const T*, scope>;
    using typename base_t::basetype;
    using typename base_t::gpu_maptype;
    using typename base_t::maptype;
    using typename base_t::reference;

    public:
    using typename base_t::value_type;

    template<typename EA = void, typename = typename enable_if<scope == memory::device, EA>::type>
    gpu_type(const T* ptr)
    {
        this->impl.reset(interface::goopax_4_gpu_type_assign(&ptr, type_enum<basetype*>::value));
        /// conversion from CPU type
    }
    gpu_type(const gpu_type& b)
    {
        /// copy constructor
        *this = b;
    }
    gpu_type(const gpu_type<T*, scope>& b)
    {
        /// conversion from pointer to const pointer.
        interface::goopax_4_gpu_type_copyassign(this->impl.get(), b.impl.get());
    }
    gpu_type()
    {}
    gpu_type& operator=(const gpu_type& b)
    {
        /// assignment operator
        interface::goopax_4_gpu_type_copyassign(this->impl.get(), b.impl.get());
        return *this;
    }

    template<typename EA = void, typename = typename enable_if<scope == memory::device, EA>::type>
    gpu_type& operator=(const T* ptr)
    {
        return *this = gpu_type(ptr);
    }
};
}
}
